import random
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List


def computed_property():
    return 42


class Animal:
    num_legs = computed_property()

    def __init__(self, name: str, age: int):
        self.name = name
        self.age = age

    def speak(self):
        raise NotImplementedError("Subclasses must implement this method")

    def __str__(self):
        return f"{self.name}, {self.age} years old"


class Dog(Animal):
    def speak(self):
        return "Woof!"


class Cat(Animal):
    def speak(self):
        return "Meow!"


class Bird(Animal):
    def speak(self):
        return "Tweet!"


def main():
    random.seed(42)  # For reproducible results
    num_animals = 10
    animals = generate_random_animals(num_animals)

    print("Animals:")
    for animal in animals:
        print(animal)

    print("\nAnimal Sounds:")
    print_animal_sounds(animals)

    oldest_animal = find_oldest_animal(animals)
    print(
        f"\nThe oldest animal is {oldest_animal.name}, who is {oldest_animal.age} years old."
    )


def generate_random_animals(num_animals: int) -> List[Animal]:
    animals = []
    for _ in range(num_animals):
        animal_type = random.choice([Dog, Cat, Bird])
        name = random.choice(["Max", "Bella", "Charlie", "Molly", "Buddy", "Luna"])
        age = random.randint(1, 15)
        animal = animal_type(name, age)
        animals.append(animal)
    return animals


def print_animal_sounds(animals: List[Animal]):
    for animal in animals:
        print(f"{animal.name} says {animal.speak()}")


def find_oldest_animal(animals: List[Animal]) -> Animal:
    oldest = max(animals, key=lambda animal: animal.age)
    return oldest


if __name__ == "__main__":
    main()



from pathlib import Path

from libcst.codemod import CodemodTest
from tato.tato import ReorderFileCodemod


class TestFilesystem(CodemodTest):
    TRANSFORM = ReorderFileCodemod

    def test_large_filesystem(self) -> None:
        self.maxDiff = None
        folder = Path(__file__).parent
        before = (folder / "before.py").read_text()
        after = (folder / "after.py").read_text()

        self.assertCodemod(before, after)



from dataclasses import dataclass, field

from tato._node import NodeType, OrderedNode


@dataclass
class Section:
    """Grouping of ordered nodes"""

    _constants: list[OrderedNode] = field(default_factory=list)
    _unknowns: list[OrderedNode] = field(default_factory=list)
    _classes: list[OrderedNode] = field(default_factory=list)
    _functions: list[OrderedNode] = field(default_factory=list)

    def add(self, node: OrderedNode) -> None:
        if node.node_type == NodeType.CONSTANT:
            self._constants.append(node)
        elif node.node_type == NodeType.UNKNOWN:
            self._unknowns.append(node)
        elif node.node_type == NodeType.CLASS:
            self._classes.append(node)
        elif node.node_type == NodeType.FUNCTION:
            # TODO: Check node.leaf_order
            self._functions.insert(0, node)
        else:
            raise ValueError(f"Unknown node_type: {node.node_type}")

    def flatten(self) -> list[OrderedNode]:
        return self._constants + self._unknowns + self._classes + self._functions


@dataclass
class SectionsBuilder:
    """Construct the sections of the output.

    A new section if we've already seen a larger node_type in the current
    section.
    """

    imports: list[OrderedNode] = field(default_factory=list)
    sections: list[Section] = field(default_factory=list)

    _largest_node_type: NodeType = field(default=NodeType.IMPORT)
    _current: Section = field(default_factory=Section)
    _sealed: bool = field(default=False)

    def add(self, node: OrderedNode) -> None:
        if self._sealed:
            raise ValueError("Cannot add to a sealed builder")
        if node.node_type == NodeType.IMPORT:
            self.imports.append(node)
            return
        if self._largest_node_type > node.node_type:
            self.sections.append(self._current)
            self._current = Section()
        self._current.add(node)
        self._largest_node_type = node.node_type

    def seal(self) -> None:
        if not self._sealed:
            self.sections.append(self._current)
            self._sealed = True



import enum
from dataclasses import dataclass
from typing import Literal, Mapping, Union

import libcst as cst

# Type of a node found in a module's body.
TopLevelNode = Union[cst.SimpleStatementLine, cst.BaseCompoundStatement]


class NodeType(enum.IntEnum):
    IMPORT = 0
    CONSTANT = 1
    UNKNOWN = 2
    CLASS = 3
    FUNCTION = 4


@dataclass(frozen=True)
class OrderedNode:
    """Information needed to order TopLevelNodes."""

    node: TopLevelNode
    node_type: NodeType
    # The ealier a TopLevelNode is accessed, the earlier it should appear in the
    # output.
    # Acesss is the (lineno, colno) in the original file.
    first_access: tuple[int, int]
    # Tie break should be the order of the node in the original file.
    prev_body_index: int
    # 'leaf_last' allows for the most important / highest abstraction to be a
    #  the top of the file.
    # Due to Python semantics, most node_type's must be 'leaf_first', except
    # functions.
    leaf_order: Literal["leaf_first", "leaf_last"]

    @classmethod
    def from_cst_node(
        cls,
        cstnode: TopLevelNode,
        first_access: Mapping[TopLevelNode, tuple[int, int]],
        indexes: Mapping[TopLevelNode, int],
    ):
        def node_type(node: TopLevelNode) -> NodeType:
            if isinstance(node, cst.SimpleStatementLine):
                if isinstance(node.body[0], (cst.Assign, cst.AnnAssign, cst.AugAssign)):
                    return NodeType.CONSTANT
                elif isinstance(node.body[0], (cst.Import, cst.ImportFrom)):
                    return NodeType.IMPORT
                else:
                    return NodeType.UNKNOWN
            elif isinstance(node, (cst.ClassDef,)):
                return NodeType.CLASS
            elif isinstance(node, (cst.FunctionDef,)):
                return NodeType.FUNCTION
            else:
                return NodeType.UNKNOWN

        def leaf_type(node: TopLevelNode) -> Literal["leaf_first", "leaf_last"]:
            if isinstance(node, (cst.FunctionDef,)):
                return "leaf_last"
            else:
                return "leaf_first"

        return cls(
            node=cstnode,
            node_type=node_type(cstnode),
            first_access=first_access[cstnode],
            prev_body_index=indexes[cstnode],
            leaf_order=leaf_type(cstnode),
        )

    def __hash__(self) -> int:
        return hash(self.node)

    def __eq__(self, other: "OrderedNode") -> bool:
        return self.node == other.node

    def __lt__(self, other: "OrderedNode") -> bool:
        return self._as_tuple() < other._as_tuple()

    def _as_tuple(self) -> tuple:
        if self.node_type == NodeType.IMPORT:
            # Try and keep imports sorted by their previous location.
            return (self.node_type, self.prev_body_index)
        else:
            if self.leaf_order == "leaf_first":
                return (self.node_type, self.first_access, self.prev_body_index)
            elif self.leaf_order == "leaf_last":
                # Err, I don't really understand why we have to multiply by -1.
                # Something about inserting functions in reverse order...
                return (
                    self.node_type,
                    (self.first_access[0] * -1, self.first_access[0] * -1),
                    self.prev_body_index,
                )
            else:
                raise ValueError(f"Unknown leaf_order: {self.leaf_order}")



from libcst.codemod import CodemodTest
from tato.tato import ReorderFileCodemod


class TestMultipass(CodemodTest):
    TRANSFORM = ReorderFileCodemod

    def test_unknown_constants_ordering(self) -> None:
        before = """
            if True:
                PEANUT = 2
            else:
                PEANUT = 2
            BUTTER = PEANUT * 1
        """
        after = """
            if True:
                PEANUT = 2
            else:
                PEANUT = 2
            BUTTER = PEANUT * 1
        """
        self.assertCodemod(before, after)

    def test_constants_unknown_ordering(self) -> None:
        before = """
            FLOWER = 1
            if True:
                WATER = FLOWER + 1
            else:
                WATER = FLOWER + 1
        """
        after = """
            FLOWER = 1
            if True:
                WATER = FLOWER + 1
            else:
                WATER = FLOWER + 1
        """
        self.assertCodemod(before, after)

    def test_function_ordering(self) -> None:
        before = """
            def second():
                pass
                
            def first():
                pass
                
            def main():
                first()
                second()
        """
        after = """
            def main():
                first()
                second()

            def first():
                pass
            def second():
                pass
        """
        self.assertCodemod(before, after)

    def test_multiple_symbols(self) -> None:
        before = """
            from x.y import z
            PEANUT = 1

            class One:
                pass
                    
            class Two(One):
                pass
                
            ALMOND = PEANUT + 1

            def a():
                b()

            def b():
                c()

            def c():
                pass
        """
        after = """
            from x.y import z
            PEANUT = 1
            
            ALMOND = PEANUT + 1

            class One:
                pass
                    
            class Two(One):
                pass
                
            def a():
                b()

            def b():
                c()

            def c():
                pass
        """

        self.assertCodemod(before, after)

    def test_mashed_potato(self) -> None:
        before = """
            import random

            def peel(vegetable):
                return vegetable[1:-1]

            def chop(vegetable):
                return list(vegetable)

            def mash(veg_pieces):
                random.shuffle(veg_pieces)
                return ''.join(veg_pieces)

            def prepare(vegetable):
                peeled = peel(vegetable)
                return chop(peeled)

            def mashed_potatoes(vegetable):
                prepared = prepare(vegetable)
                return mash(prepared)

            if __name__ == "__main__":
                potato = ' potato '
                print(mashed_potatoes(potato))
        """
        after = """
            import random

            def mashed_potatoes(vegetable):
                prepared = prepare(vegetable)
                return mash(prepared)

            def prepare(vegetable):
                peeled = peel(vegetable)
                return chop(peeled)

            def mash(veg_pieces):
                random.shuffle(veg_pieces)
                return ''.join(veg_pieces)

            def peel(vegetable):
                return vegetable[1:-1]

            def chop(vegetable):
                return list(vegetable)

            if __name__ == "__main__":
                potato = ' potato '
                print(mashed_potatoes(potato))
        """
        self.assertCodemod(before, after)



# Modified from https://github.com/google/importlab/blob/9751bfc27d5bde83b117c85eee48cb9b6cc0d652/importlab/fs.py
import abc
import glob
import os
import tarfile
import tempfile


class FileSystemError(Exception):
    pass


class FileSystem(abc.ABC):
    """Interface for file systems."""

    @abc.abstractmethod
    def isfile(self, path):
        """Is this a file?"""
        pass

    @abc.abstractmethod
    def isdir(self, path):
        """Is this a directory?"""
        pass

    @abc.abstractmethod
    def read(self, path):
        """Read a file."""
        pass

    @abc.abstractmethod
    def refer_to(self, path):
        """Get a fully qualified path for the given path."""
        pass

    def relative_path(self, path):
        """Return the relative path to `path`.

        If this filesystem has a root directory, and path is within that
        directory tree, return the relative path; otherwise return None.
        """
        return None


class StoredFileSystem(FileSystem):
    """File system based on a file list."""

    def __init__(self, files):
        self.files = files
        self.dirs = {os.path.dirname(f) for f in files}

    def isfile(self, path):
        return path in self.files

    def isdir(self, path):
        return path in self.dirs

    def read(self, path):
        return self.files[path]

    def refer_to(self, path):
        return path


class OSFileSystem(FileSystem):
    """File system that uses an OS file system underneath."""

    def __init__(self, root):
        assert root is not None
        self.root = root
        _, tmp_path = tempfile.mkstemp()
        self._is_case_insensitive = os.path.exists(tmp_path.upper())

    def _join(self, path):
        return os.path.join(self.root, path)

    def _matches_path(self, path):
        if self._is_case_insensitive:
            return path in glob.glob(path + "*")
        return True

    def isfile(self, path):
        assert path is not None
        fullpath = self._join(path)
        return os.path.isfile(fullpath) and self._matches_path(fullpath)

    def isdir(self, path):
        assert path is not None
        fullpath = self._join(path)
        return os.path.isdir(fullpath) and self._matches_path(fullpath)

    def read(self, path):
        with open(self._join(path), "r") as fi:
            return fi.read()

    def refer_to(self, path):
        return self._join(path)

    def relative_path(self, path):
        if path.startswith(self.root):
            return path[len(self.root) + 1 :]
        return None


class RemappingFileSystem(FileSystem, abc.ABC):
    """File system wrapper that transforms a path before looking it up."""

    def __init__(self, underlying):
        self.underlying = underlying

    @abc.abstractmethod
    def map_path(self, path):
        pass

    def isfile(self, path):
        return self.underlying.isfile(self.map_path(path))

    def isdir(self, path):
        return self.underlying.isdir(self.map_path(path))

    def read(self, path):
        return self.underlying.read(self.map_path(path))

    def refer_to(self, path):
        return self.underlying.refer_to(self.map_path(path))


class ExtensionRemappingFileSystem(RemappingFileSystem):
    """File system that remaps .py file extensions."""

    def __init__(self, underlying, extension):
        super(ExtensionRemappingFileSystem, self).__init__(underlying)
        self.extension = extension

    def map_path(self, path):
        p, ext = os.path.splitext(path)
        if ext == ".py":
            return p + "." + self.extension
        return path


class PYIFileSystem(ExtensionRemappingFileSystem):
    """File system that remaps .py file extensions to pyi."""

    def __init__(self, underlying):
        super(PYIFileSystem, self).__init__(underlying, "pyi")


class TarFileSystem:
    """Filesystem that serves files out of a .tar."""

    def __init__(self, tar):
        self.tar = tar
        self.files = list(t.name for t in tar.getmembers() if t.isfile())
        self.directories = list(t.name for t in tar.getmembers() if t.isdir())
        self.top_level = {f.split(os.path.sep)[0] for f in self.files}

    def isfile(self, path):
        return any(os.path.join(top, path) in self.files for top in self.top_level)

    def isdir(self, path):
        return any(os.path.join(top, path) in self.files for top in self.top_level)

    def read(self, path):
        return self.tar.extractfile(path).read()

    def refer_to(self, path):
        return "tar:" + path

    @staticmethod
    def read_tarfile(archive_filename):
        tar = tarfile.open(archive_filename)
        return TarFileSystem(tar)


class Path(object):
    def __init__(self, paths=None):
        self.paths = paths if paths else []

    def add_path(self, path, kind="os"):
        if kind == "os":
            path = OSFileSystem(path)
        elif kind == "pyi":
            path = PYIFileSystem(OSFileSystem(path))
        else:
            raise FileSystemError("Unrecognized filesystem type: ", kind)
        self.paths.append(path)

    def add_fs(self, fs):
        assert isinstance(fs, FileSystem), "Unrecognised filesystem: %r" % fs
        self.paths.append(fs)



import sys

import libcst.tool
from libcst._version import __version__ as libcst_version

from tato.__about__ import __version__


def main() -> None:
    if len(sys.argv) < 2 or sys.argv[1] in ("-h", "--help"):
        print("usage: tato [codemod] PATH [PATH ...]")
        return
    if sys.argv[1] in ("--version",):
        print(f"tato version {__version__} (libcst version {libcst_version})")
        return

    # The help text from libcst spits out 'usage: tato codemod' and exposes the
    # underlying libcst configuration. We can reuse that for now.
    args = sys.argv[1:]
    if sys.argv[1] == "codemod":
        args = sys.argv[2:]
    sys.exit(
        libcst.tool.main(
            "tato",
            ["codemod", "-x", "tato.tato.ReorderFileCodemod"] + args,
        )
    )



import heapq
from collections import defaultdict
from typing import Mapping, cast

import libcst as cst
from libcst import codemod
from libcst.metadata import (
    Assignment,
    CodeRange,
    ParentNodeProvider,
    PositionProvider,
    ProviderT,
    Scope,
    ScopeProvider,
)

from tato._node import NodeType, OrderedNode, TopLevelNode
from tato._section import Section, SectionsBuilder

# Expected to be larger than any possible line number.
LARGE_NUM = 10_000_000


class ReorderFileCodemod(codemod.VisitorBasedCodemodCommand):
    """Reorder files based on their internal file dependencies.

    Generally, it tries to order files by:
        1. Imports
        2. Constants/Expressions
        3. Classes (leaf classes first)
        4. Functions (leaf functions last)

    Ocassionally, due to Python semantics, we'll have to break this order.
    In that case, we'll organize the file into sections, where each section
    is ordered.
    """

    METADATA_DEPENDENCIES = (
        ScopeProvider,
        ParentNodeProvider,
        PositionProvider,
    )

    def should_allow_multiple_passes(self) -> bool:
        """
        Call transform repeatedly until the tree doesn't change between passes.

        `test_mashed_potato()` shows a case when this necessary.
        """
        return True

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        graph = create_graph(original_node, self.metadata)
        topo_sorted = topological_sort(graph)
        imports, sections = categorize_nodes(topo_sorted)

        return updated_node.with_changes(
            body=[i.node for i in imports]
            + [n.node for s in sections for n in s.flatten()]
        )


def create_graph(
    module: cst.Module, metadata: Mapping[ProviderT, Mapping[cst.CSTNode, object]]
) -> dict[OrderedNode, set[OrderedNode]]:
    """Create a graph of definitions (assignments)

    :: returns:
        A tuple of two dicts:
            - A graph mapping between nodes (where nodes are top-level
              objects in a module's body)
            - A mapping of definitions to their first access position.

    Example:
        ```
        1 def a():
        2     b()
        3     c()
        4
        5 def b():
        6     c()
        7
        8 def c(): pass
        ```
        will return a graph:
            `{'a': {'b', 'c'}, 'b': {'c'}, 'c': set()}`
        and first access mapping:
            `{a: (10_000_000, 10_000_000), b: (2, 4), c: (3, 4)}`

    """
    scopes = cast(Mapping[cst.CSTNode, Scope], metadata[ScopeProvider]).values()
    parents = cast(Mapping[cst.CSTNode, cst.CSTNode], metadata[ParentNodeProvider])
    positions = cast(Mapping[cst.CSTNode, CodeRange], metadata[PositionProvider])

    modulebodyset: set[TopLevelNode] = set(module.body)
    globalscope = next((s.globals for s in scopes if s is not None))

    def find_top_level_node(node: cst.CSTNode) -> TopLevelNode:
        """Find the `cst.Module.body` that contains the given node."""
        while node not in modulebodyset:
            node = parents[node]
        return cast(TopLevelNode, node)

    graph: dict[TopLevelNode, list[TopLevelNode]] = {}
    first_access: dict[TopLevelNode, tuple[int, int]] = defaultdict(
        lambda: (LARGE_NUM, LARGE_NUM)
    )
    for node in module.body:
        graph[node] = []

    for scope in scopes:
        if scope is None:
            continue
        for access in scope.accesses:
            # Using first global assignment. This seems likely to be buggy.
            globalassignment = next(
                (
                    a
                    for a in access.referents
                    if a.scope == globalscope and isinstance(a, Assignment)
                ),
                None,
            )
            # Skip non-global assignments.
            if not globalassignment:
                continue

            top_level_assignment = find_top_level_node(globalassignment.node)
            top_level_access = find_top_level_node(access.node)

            # Skip self-edges.
            if top_level_assignment == top_level_access:
                continue

            # Create edge from top_level_assignment to top_level_access
            graph[top_level_assignment].append(top_level_access)

            # Track first access of the assignment.
            coderange = positions[access.node]
            first_access[top_level_assignment] = min(
                first_access[top_level_assignment],
                (coderange.start.line, coderange.start.column),
            )

    indexes = {node: i for i, node in enumerate(module.body)}
    ordered_nodes = [
        OrderedNode.from_cst_node(node, first_access, indexes) for node in module.body
    ]
    lookup = {n.node: n for n in ordered_nodes}
    return {lookup[k]: set(lookup[v] for v in vs) for k, vs in graph.items()}


def topological_sort(
    graph: dict[OrderedNode, set[OrderedNode]],
) -> list[OrderedNode]:
    """
    Sorts a graph of definitions into a topological order.

    Example:
    >>> topological_sort({'a': {'b'}, 'b': {'c'}, 'c': set(), 'd': set()})
    ['d', 'c', 'b', 'a']
    """

    topo_sorted = []
    innodes: defaultdict[OrderedNode, int] = defaultdict(int)
    for src, dsts in graph.items():
        innodes[src]
        # Ignore usages of imports when sorting.
        if src.node_type == NodeType.IMPORT:
            continue
        for dst in dsts:
            innodes[dst] += 1

    # Using a heap (sorted list) ensures each section is ordered and each time
    # we see something out of order, we can start a new section.
    heap = [node for node, count in innodes.items() if count == 0]
    heapq.heapify(heap)
    while heap:
        node = heapq.heappop(heap)
        topo_sorted.append(node)
        for dst in graph[node]:
            innodes[dst] -= 1
            if innodes[dst] == 0:
                heapq.heappush(heap, dst)
    return topo_sorted


def categorize_nodes(
    topo_sorted: list[OrderedNode],
) -> tuple[list[OrderedNode], list[Section]]:
    """Categorize into: imports + sections.

    A section is:
     - constants
     - unknonws
     - classes
     - functions

    We start a new section when something is "out of order", so each section is
    ordered (constants, unknowns, classes, functions).
    """
    builder = SectionsBuilder()

    for node in topo_sorted:
        builder.add(node)
    builder.seal()

    return builder.imports, builder.sections


def _print_code(node: cst.CSTNode) -> None:
    """Print the code of a node. Used for debugging."""
    tree = cst.parse_module("")
    tree = tree.with_changes(body=[node])
    print(tree.code)
